# VTK-Based visualiser
# intended to replace realtime_visualisation_new.py

import threading
import Tkinter
import vtk
from vtk.tk.vtkTkRenderWidget import vtkTkRenderWidget

class Visualiser(threading.Thread):

    """A VTK-powered visualiser, designed to replace the VPython one.
    Intended to be run in its own thread

    Functions:

    run: Overrides threading.Thread.run(). Do not call. Called by
    inherited start().

    initialise_gui: Create the TK window and makes the vtk renderer
    available.

    add_axes: Draw axes onto the renderer.
    
    """

    def __init__(self, domain, scale_z=1.0, rect=None, title='Test'):
        threading.Thread.__init__(self)
        # Initialise data structures. coloring, setup and updating are maps
        # quantity name -> boolean, setup means to render it,
        # update means to update with time.
        # qcolor maps quantity name -> (float, float, float): the colour
        # to render a quantity.
        self.setup = {}
        self.updating = {}
        self.coloring = {}
        self.qcolor = {}
        self.scale_z = {}
        self.default_scale_z = scale_z
        self.domain = domain
        self.vertices = domain.vertex_coordinates

        # Default options
        for x in self.domain.quantities:
            self.setup[x] = False
            self.updating[x] = False
            self.qcolor[x] = (0.0, 0.0, 0.0)
            self.coloring[x] = False

        # Bounding box.
        if rect is None:
            self.max_x = max(max(self.vertices[:,0]),max(self.vertices[:,2]),max(self.vertices[:,4]))
            self.min_x = min(min(self.vertices[:,0]),min(self.vertices[:,2]),min(self.vertices[:,4]))
            self.max_y = max(max(self.vertices[:,1]),max(self.vertices[:,3]),max(self.vertices[:,5]))
            self.min_y = min(min(self.vertices[:,1]),min(self.vertices[:,3]),min(self.vertices[:,5]))
        else:
            self.max_x = rect[2]
            self.min_x = rect[0]
            self.max_y = rect[3]
            self.min_y = rect[1]

        self.range_x = self.max_x - self.min_x
        self.range_y = self.max_y - self.min_y
        self.range_xy = max(self.range_x, self.range_y)
        # Start this off in its own thread.
        self.start()
        
    def run(self):
        self.initialise_gui()
        self.add_axes()
        self.setup_all()
        self.root.mainloop()

    def initialise_gui(self):
        """Prepare the GUI for the Visualiser, and set up the
        renderer.
        
        """
        # Using the TK VTK widget allows extendability
        # should a gui need adding.
        self.root = Tkinter.Tk()
        self.renderWidget = vtkTkRenderWidget(self.root, width=400, height=400)
        self.renderWidget.pack(expand='true', fill='both')
        self.renderWindow = self.renderWidget.GetRenderWindow()
        self.renderer = vtk.vtkRenderer()
        self.renderWindow.AddRenderer(self.renderer)

    def add_axes(self):
        """Add axes to this Visualiser
        """
        pass

    def setup_all(self):
        """Draw in the data that is specified by setup or update
        """

        self.N_tri = len(self.domain.triangles)
        self.grids = {}
        self.cells = vtk.vtkCellArray()
        self.vertices = self.domain.get_vertex_coordinates()
        self.vert_index = {}

        for n in range(len(self.vertices)):
            for i in range(3):
                self.vert_index[self.domain.triangles[n][i]] = self.vertices[n][i*2:i*2+2]

        self.N_vert = len(self.vert_index.keys())

        # Prepare the list of cells
        for t in self.domain.triangles:
            self.cells.InsertNextCell(3)
            for i in range(3):
                self.cells.InsertCellPoint(t[i])

        # Set up the rendering of each quantity
        for q in self.domain.quantities:
            if self.setup[q] | self.updating[q]:
                self.qty_index = {}
                self.draw_quantity(q)

    def draw_quantity(self, q):
        for n in range(self.N_tri):
            for v in range(3):
                self.qty_index[self.domain.triangles[n][v]] = self.domain.quantities[q].vertex_values[n][v]
                
        self.points = vtk.vtkPoints()
        for v in range(self.N_vert):
            self.points.InsertNextPoint(self.vert_index[v][0],
                                        self.vert_index[v][1],
                                        self.qty_index[v])

        self.grid = vtk.vtkPolyData()
        self.grid.SetPoints(self.points)
        self.grid.SetPolys(self.cells)
        self.grids[q] = self.grid

        self.mapper = vtk.vtkPolyDataMapper()
        self.mapper.SetInput(self.grid)

        self.actor = vtk.vtkActor()
        self.actor.SetMapper(self.mapper)
        if self.coloring[q] is True:
            self.actor.GetProperty().SetColor(self.qcolor[q])
        else:
            self.actor.GetProperty().SetColor(0.5, 0.5, 0.5)

        print "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
        self.renderer.AddActor(self.actor)
        print "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"

    def update_all(self):
        pass

    def update_timer(self):
        pass
    
if __name__ == '__main__':
    v = Visualiser()
    v.start()
    
